"""
Unittests of asm2cfg's regexes
"""

import unittest

from src.asm2cfg import asm2cfg


class FunctionHeaderTestCase(unittest.TestCase):
    """
    Tests of function header regex
    """

    def test_gdb_unstripped(self):
        line = 'Dump of assembler code for function test_function:'
        fmt, fun = asm2cfg.parse_function_header(line)

        self.assertEqual(fmt, asm2cfg.InputFormat.GDB)
        self.assertEqual(fun, 'test_function')

    def test_gdb_stripped(self):
        line = 'Dump of assembler code from 0x555555555faf to 0x555555557008:'
        fmt, fun = asm2cfg.parse_function_header(line)

        self.assertEqual(fmt, asm2cfg.InputFormat.GDB)
        self.assertEqual(fun, '0x555555555faf-0x555555557008')

    def test_objdump(self):
        line = '000000000000100b <bar>:'
        fmt, fun = asm2cfg.parse_function_header(line)

        self.assertEqual(fmt, asm2cfg.InputFormat.OBJDUMP)
        self.assertEqual(fun, 'bar')


class ParseAddressTestCase(unittest.TestCase):
    """
    Tests of parse_address function
    """

    def test_absolute(self):
        line = '0x000055555557259c:	XYZ'
        address, rest = asm2cfg.parse_address(line)

        self.assertIsNot(address, None)
        self.assertEqual(address.abs, 0x55555557259c)
        self.assertIs(address.base, None)
        self.assertIs(address.offset, None)
        self.assertEqual(rest, '	XYZ')

    def test_relative(self):
        line = '0x000055555557259c <+11340>:	XYZ'
        address, rest = asm2cfg.parse_address(line)

        self.assertIsNot(address, None)
        self.assertEqual(address.abs, 0x55555557259c)
        self.assertIs(address.base, None)
        self.assertEqual(address.offset, 11340)
        self.assertEqual(rest, '	XYZ')


class ParseBodyTestCase(unittest.TestCase):
    """
    Tests of asm2cfg.parse_body function
    """

    def setUp(self):
        self.target_info = asm2cfg.X86TargetInfo()

    def test_gdb_stripped_known(self):
        line = '	call 0x55555558add0 <_Z19exportDebugifyStats>'
        body, opcode, ops, rest = asm2cfg.parse_body(line, self.target_info)

        self.assertIsNot(body, None)
        self.assertEqual(body, 'call 0x55555558add0')
        self.assertEqual(opcode, 'call')
        self.assertEqual(ops, ['0x55555558add0'])
        self.assertEqual(rest, '<_Z19exportDebugifyStats>')

    def test_gdb_stripped_pic(self):
        line = '    call   *0x26a16(%rip)        # 0x5555555967a8'
        body, opcode, ops, rest = asm2cfg.parse_body(line, self.target_info)

        self.assertIsNot(body, None)
        self.assertEqual(body, 'call   *0x26a16(%rip)')
        self.assertEqual(opcode, 'call')
        self.assertEqual(ops, ['*0x26a16(%rip)'])
        self.assertEqual(rest, '# 0x5555555967a8')

    def test_gdb_plt(self):
        line = '	callq  0x1020 <foo@plt>'
        body, opcode, ops, rest = asm2cfg.parse_body(line, self.target_info)

        self.assertIsNot(body, None)
        self.assertEqual(body, 'callq  0x1020')
        self.assertEqual(opcode, 'callq')
        self.assertEqual(ops, ['0x1020'])
        self.assertEqual(rest, '<foo@plt>')

    def test_gdb_stripped_nonpic(self):
        line = '	call   0x555555555542'
        body, opcode, ops, rest = asm2cfg.parse_body(line, self.target_info)

        self.assertIsNot(body, None)
        self.assertEqual(body, 'call   0x555555555542')
        self.assertEqual(opcode, 'call')
        self.assertEqual(ops, ['0x555555555542'])
        self.assertEqual(rest, '')

    def test_gdb_indirect_call(self):
        line = '	callq  *(%rsi)'
        body, opcode, ops, rest = asm2cfg.parse_body(line, self.target_info)

        self.assertIsNot(body, None)
        self.assertEqual(body, 'callq  *(%rsi)')
        self.assertEqual(opcode, 'callq')
        self.assertEqual(ops, ['*(%rsi)'])
        self.assertEqual(rest, '')


class ParseTargetTestCase(unittest.TestCase):
    """
    Tests of parse_address function
    """

    def test_with_offset(self):
        line = '<_Z19exportDebugifyStats+123>'
        address, rest = asm2cfg.parse_target(line)

        self.assertIsNot(address, None)
        self.assertIs(address.abs, None)
        self.assertEqual(address.base, '_Z19exportDebugifyStats')
        self.assertEqual(address.offset, 123)
        self.assertEqual(rest, '')

    def test_with_neg_offset(self):
        line = '<_Z19exportDebugifyStats-123>'
        address, rest = asm2cfg.parse_target(line)

        self.assertIsNot(address, None)
        self.assertIs(address.abs, None)
        self.assertEqual(address.base, '_Z19exportDebugifyStats')
        self.assertEqual(address.offset, -123)
        self.assertEqual(rest, '')

    def test_without_offset(self):
        line = '<_Z19exportDebugifyStats>'
        address, rest = asm2cfg.parse_target(line)

        self.assertIsNot(address, None)
        self.assertIs(address.abs, None)
        self.assertEqual(address.base, '_Z19exportDebugifyStats')
        self.assertEqual(address.offset, 0)
        self.assertEqual(rest, '')


class ParseCommentTestCase(unittest.TestCase):
    """
    Tests of parse_comment function
    """

    def setUp(self):
        self.target_info = asm2cfg.X86TargetInfo()

    def test_absolute(self):
        line = '# 0x5555555967a8'
        address, rest = asm2cfg.parse_comment(line, self.target_info)

        self.assertIsNot(address, None)
        self.assertEqual(address.abs, 0x5555555967a8)
        self.assertIs(address.base, None)
        self.assertIs(address.offset, None)
        self.assertEqual(rest, '')

    def test_symbolic(self):
        line = '# 0x5555555967a8 <foo>'
        address, rest = asm2cfg.parse_comment(line, self.target_info)

        self.assertIsNot(address, None)
        self.assertEqual(address.abs, 0x5555555967a8)
        self.assertEqual(address.base, 'foo')
        self.assertIs(address.offset, 0)
        self.assertEqual(rest, '')

    def test_complete(self):
        line = '# 3ff8 <foo+0x2ff8>'
        address, rest = asm2cfg.parse_comment(line, self.target_info)

        self.assertIsNot(address, None)
        self.assertEqual(address.abs, 0x3ff8)  # FIXME: support hex offsets
        self.assertEqual(address.base, 'foo')
        self.assertEqual(address.offset, 0x2ff8)
        self.assertEqual(rest, '')



"""
This file includes the GDB extension for asm2cfg. Commands and settings are
exposed to GDB by extending the certain classes.
For further information see
https://sourceware.org/gdb/current/onlinedocs/gdb/Python.html#Python.
"""


import traceback

import gdb

from asm2cfg import asm2cfg


class SkipCalls(gdb.Parameter):
    """Manage skipcalls parameter.

    Usage: set skipcalls
           set skipcalls off
    """

    set_doc = 'Set whether savecfg and viewcfg commands will skip function calls from splitting CFG blocks'
    show_doc = 'Set whether savecfg and viewcfg commands will skip function calls from splitting CFG blocks'

    def __init__(self):
        super().__init__('skipcalls', gdb.COMMAND_DATA, gdb.PARAM_BOOLEAN)
        self.value = False

    def get_set_string(self):
        return f'Commands savecfg and viewcfg will skip function calls \
                from splitting CFG blocks: {self.value_to_string()}'

    def get_show_string(self, _):
        return f'Commands savecfg and viewcfg will skip function calls \
                from splitting CFG blocks: {self.value_to_string()}'

    def value_to_string(self):
        if self.value:
            return 'on'
        return 'off'


class ViewCfg(gdb.Command):  # pylint: disable=too-few-public-methods
    """
    Draw an assembly control-flow graph (CFG) of the currently executed
    function. If function is big and CFG rendering takes too long, try to
    skip function calls from splitting the code with 'set skipcalls on'. Or
    save the graph with 'savecfg' command and view it with other program.'
    """

    def __init__(self):
        super().__init__('viewcfg', gdb.COMMAND_USER)

    def invoke(self, _arg, _from_tty):  # pylint: disable=no-self-use
        """ Called by GDB when viewcfg command is invoked """
        try:
            frame = gdb.selected_frame()
            arch = frame.architecture().name()
            if arch.startswith('i386'):
                target_name = 'x86'
            elif arch.startswith('arm'):
                target_name = 'arm'
            else:
                raise RuntimeError(f'unknown platform: {arch}')
            assembly_lines = gdb.execute('disassemble', from_tty=False, to_string=True).split('\n')
            function_name, basic_blocks = asm2cfg.parse_lines(assembly_lines, gdb.parameter('skipcalls'),
                                                              target_name)
            asm2cfg.draw_cfg(function_name, basic_blocks, view=True)
        # Catch error coming from GDB side before other errors.
        except gdb.error as ex:
            raise gdb.GdbError(ex)
        except Exception as ex:
            traceback.print_exc()
            raise gdb.GdbError(ex)


class SaveCfg(gdb.Command):  # pylint: disable=too-few-public-methods
    """
    Save an assembly control-flow graph (CFG) of the currently executed
    function. If function is big and CFG rendering takes too long, try to
    skip function calls from splitting the code with 'set skipcalls on'.
    """

    def __init__(self):
        super().__init__('savecfg', gdb.COMMAND_USER)

    def invoke(self, _arg, _from_tty):  # pylint: disable=no-self-use
        """ Called by GDB when savecfg command is invoked """
        try:
            assembly_lines = gdb.execute('disassemble', from_tty=False, to_string=True).split('\n')
            function_name, basic_blocks = asm2cfg.parse_lines(assembly_lines, gdb.parameter('skipcalls'),
                                                              'x86')
            asm2cfg.draw_cfg(function_name, basic_blocks, view=False)
        # Catch error coming from GDB side before other errors.
        except gdb.error as ex:
            raise gdb.GdbError(ex)
        except Exception as ex:
            traceback.print_exc()
            raise gdb.GdbError(ex)


# Instantiate the settings and commands.
SkipCalls()
ViewCfg()
SaveCfg()



"""
Unittests of asm2cfg's assembly parser
"""

import unittest

from src.asm2cfg import asm2cfg


def _get_the_source_block(blocks):
    sources = [block for _, block in blocks.items() if block.jump_edge is not None]
    return sources[0] if len(sources) == 1 else None


class ParseLineTestCase(unittest.TestCase):
    """
    Tests of parse_line function
    """

    def setUp(self):
        self.target_info = asm2cfg.X86TargetInfo()
        self.arm_target_info = asm2cfg.ARMTargetInfo()

    def test_simple_inst(self):
        line = '0x000055555556f957 <+7>:	push   %r14'
        i = asm2cfg.parse_line(line, 1, 'main', asm2cfg.InputFormat.GDB, self.target_info)

        self.assertIsNot(i, None)
        self.assertEqual(i.body, 'push   %r14')
        self.assertEqual(i.lineno, 1)
        self.assertIsNot(i.address, None)
        self.assertEqual(i.address.abs, 0x000055555556f957)
        self.assertIs(i.address.base, 'main')
        self.assertEqual(i.address.offset, 7)
        self.assertIs(i.target, None)

        self.assertFalse(i.is_call())
        self.assertFalse(i.is_jump())
        self.assertFalse(i.is_unconditional_jump())

    def test_jump(self):
        line = '''\
0x00007ffff7fbf26b <+395>:	jmp    0x7ffff7fbf55d <test_function+1149>\
'''
        i = asm2cfg.parse_line(line, 1, 'main', asm2cfg.InputFormat.GDB, self.target_info)
        self.assertIsNot(i, None)

        self.assertEqual(i.body, 'jmp    0x7ffff7fbf55d')
        self.assertEqual(i.lineno, 1)
        self.assertIsNot(i.address, None)
        self.assertEqual(i.address.abs, 0x00007ffff7fbf26b)
        self.assertIs(i.address.base, 'main')
        self.assertEqual(i.address.offset, 395)
        self.assertIsNot(i.target, None)
        self.assertIs(i.target.abs, None)  # FIXME
        self.assertEqual(i.target.base, 'test_function')
        self.assertEqual(i.target.offset, 1149)

        self.assertFalse(i.is_call())
        self.assertTrue(i.is_jump())
        self.assertTrue(i.is_unconditional_jump())

    def test_branch(self):
        line = '''\
0x00007ffff7fbf565 <+1157>:	je     0x7ffff7fbf635 <test_function+1365>\
'''
        i = asm2cfg.parse_line(line, 1, 'main', asm2cfg.InputFormat.GDB, self.target_info)

        self.assertIsNot(i, None)
        self.assertEqual(i.body, 'je     0x7ffff7fbf635')
        self.assertEqual(i.lineno, 1)
        self.assertIsNot(i.address, None)
        self.assertEqual(i.address.abs, 0x00007ffff7fbf565)
        self.assertIs(i.address.base, 'main')
        self.assertEqual(i.address.offset, 1157)
        self.assertIsNot(i.target, None)
        self.assertIs(i.target.abs, None)  # FIXME
        self.assertEqual(i.target.base, 'test_function')
        self.assertEqual(i.target.offset, 1365)

        self.assertFalse(i.is_call())
        self.assertTrue(i.is_jump())
        self.assertFalse(i.is_unconditional_jump())

    def test_call(self):
        line = '''\
0x000000000002ec0f <+63>:	callq  0x2eab0 <__sigsetjmp@plt>
'''
        i = asm2cfg.parse_line(line, 1, 'main', asm2cfg.InputFormat.GDB, self.target_info)

        self.assertIsNot(i, None)
        self.assertEqual(i.body, 'callq  0x2eab0')
        self.assertEqual(i.lineno, 1)
        self.assertIsNot(i.address, None)
        self.assertEqual(i.address.abs, 0x000000000002ec0f)
        self.assertIs(i.address.base, 'main')
        self.assertEqual(i.address.offset, 63)
        self.assertIsNot(i.target, None)
        self.assertIs(i.target.abs, None)  # FIXME
        self.assertEqual(i.target.base, '__sigsetjmp@plt')
        self.assertEqual(i.target.offset, 0)

        self.assertTrue(i.is_call())
        self.assertFalse(i.is_jump())
        self.assertFalse(i.is_unconditional_jump())

    def test_call_stripped(self):
        line = '''\
 0x000055555556f9b0 <+96>:	call   *0x2731a(%rip)        # 0x555555596cd0
'''
        i = asm2cfg.parse_line(line, 1, 'main', asm2cfg.InputFormat.GDB, self.target_info)

        self.assertIsNot(i, None)
        self.assertEqual(i.body, 'call   *0x2731a(%rip)')
        self.assertEqual(i.lineno, 1)
        self.assertIsNot(i.address, None)
        self.assertEqual(i.address.abs, 0x000055555556f9b0)
        self.assertIs(i.address.base, 'main')
        self.assertEqual(i.address.offset, 96)
        self.assertIsNot(i.target, None)
        self.assertEqual(i.target.abs, 0x555555596cd0)
        self.assertIs(i.target.base, None)
        self.assertIs(i.target.offset, None)

        self.assertTrue(i.is_call())
        self.assertFalse(i.is_jump())
        self.assertFalse(i.is_unconditional_jump())

    def test_objdump(self):
        line = '''\
16bbb:	74 29                	je     16be6 <_obstack_allocated_p@@Base+0x36>
'''
        i = asm2cfg.parse_line(line, 1, 'main', asm2cfg.InputFormat.OBJDUMP, self.target_info)

        self.assertIsNot(i, None)
        self.assertEqual(i.body, 'je     16be6')
        self.assertEqual(i.lineno, 1)
        self.assertIsNot(i.address, None)
        self.assertEqual(i.address.abs, 0x16bbb)
        self.assertIs(i.address.base, 'main')
        self.assertIsNot(i.target, None)
        self.assertIs(i.target.abs, None)
        self.assertEqual(i.target.base, '_obstack_allocated_p@@Base')
        self.assertIs(i.target.offset, 0x36)

        self.assertTrue(i.is_jump())
        self.assertFalse(i.is_unconditional_jump())

    def test_arm_branch(self):
        line = '''\
  1c:	0a000001 	beq	28 <check_one_fd+0x28>
'''
        i = asm2cfg.parse_line(line, 1, 'main', asm2cfg.InputFormat.OBJDUMP, self.arm_target_info)

        self.assertIsNot(i, None)
        self.assertEqual(i.body, 'beq	28')
        self.assertEqual(i.lineno, 1)
        self.assertIsNot(i.address, None)
        self.assertEqual(i.address.abs, 0x1c)
        self.assertIs(i.address.base, 'main')
        self.assertIsNot(i.target, None)
        self.assertIs(i.target.abs, None)
        self.assertEqual(i.target.base, 'check_one_fd')
        self.assertIs(i.target.offset, 0x28)

        self.assertTrue(i.is_jump())
        self.assertFalse(i.is_unconditional_jump())

    def test_arm_jump(self):
        line = '''\
  1c:	0a000001 	b	28 <check_one_fd+0x28>
'''
        i = asm2cfg.parse_line(line, 1, 'main', asm2cfg.InputFormat.OBJDUMP, self.arm_target_info)

        self.assertIsNot(i, None)
        self.assertEqual(i.body, 'b	28')
        self.assertEqual(i.lineno, 1)
        self.assertIsNot(i.address, None)
        self.assertEqual(i.address.abs, 0x1c)
        self.assertIs(i.address.base, 'main')
        self.assertIsNot(i.target, None)
        self.assertIs(i.target.abs, None)
        self.assertEqual(i.target.base, 'check_one_fd')
        self.assertIs(i.target.offset, 0x28)

        self.assertTrue(i.is_jump())
        self.assertTrue(i.is_unconditional_jump())


class ParseLinesTestCase(unittest.TestCase):
    """
    Tests of parse_lines function
    """

    def test_linear_sequence(self):
        lines = '''\
Dump of assembler code for function main:
   0x000055555556f952 <+2>:	mov    $0x1,%ecx
   0x000055555556f957 <+7>:	push   %r14
   0x000055555556f959 <+9>:	push   %r13
   0x000055555556f95b <+11>:	push   %r12
   0x000055555556f95d <+13>:	push   %rbp
   0x000055555556f95e <+14>:	push   %rbx\
'''.split('\n')
        _, blocks = asm2cfg.parse_lines(lines, False, 'x86')

        self.assertEqual(len(blocks), 1)
        _, block = blocks.popitem()
        self.assertEqual(len(block.instructions), 6)
        self.assertIs(block.jump_edge, None)
        self.assertIs(block.no_jump_edge, None)

    def test_unconditional(self):
        lines = '''\
Dump of assembler code for function main:
   0x000055555556fffb <+1707>:	jmp    0x555555570058 <main+1800>
   0x0000555555570058 <+1800>:	mov    0xe0(%rsp),%rdi
   0x0000555555570060 <+1808>:	test   %rdi,%rdi
'''.split('\n')
        _, blocks = asm2cfg.parse_lines(lines, False, 'x86')

        self.assertEqual(len(blocks), 2)

        source_block = _get_the_source_block(blocks)
        self.assertIsNot(source_block.jump_edge, None)
        self.assertIs(source_block.no_jump_edge, None)
        self.assertEqual(len(source_block.instructions), 1)

        dst_block = blocks[source_block.jump_edge]
        self.assertIs(dst_block.jump_edge, None)
        self.assertIs(dst_block.no_jump_edge, None)
        self.assertEqual(len(dst_block.instructions), 2)

    def test_conditional(self):
        lines = '''\
Dump of assembler code for function main:
   0x000055555556fffb <+1707>:	je    0x555555570058 <main+1800>
   0x000055555556fffd <+1709>:	push   %rbx
   0x000055555556fffe <+1710>:	mov    %r15,%r8
   0x000055555556fffe <+1710>:	mov    %r15,%r8
   0x0000555555570058 <+1800>:	mov    0xe0(%rsp),%rdi
   0x0000555555570060 <+1808>:	test   %rdi,%rdi
'''.split('\n')
        _, blocks = asm2cfg.parse_lines(lines, False, 'x86')

        self.assertEqual(len(blocks), 3)

        source_block = _get_the_source_block(blocks)
        self.assertIsNot(source_block.jump_edge, None)
        self.assertIsNot(source_block.no_jump_edge, None)
        self.assertEqual(len(source_block.instructions), 1)

        fall_block = blocks[source_block.no_jump_edge]
        self.assertIs(fall_block.jump_edge, None)
        self.assertIsNot(fall_block.no_jump_edge, None)
        self.assertEqual(len(fall_block.instructions), 3)

        dst_block = blocks[source_block.jump_edge]
        self.assertIs(dst_block.jump_edge, None)
        self.assertIs(dst_block.no_jump_edge, None)
        self.assertEqual(len(dst_block.instructions), 2)

    def test_return(self):
        lines = '''\
Dump of assembler code for function main:
   0x000055555556fffb <+1707>:	retq
   0x0000555555570058 <+1800>:	mov    0xe0(%rsp),%rdi
   0x0000555555570060 <+1808>:	test   %rdi,%rdi
'''.split('\n')
        _, blocks = asm2cfg.parse_lines(lines, False, 'x86')

        self.assertEqual(len(blocks), 2)

        ret_block, other_block = blocks.values()

        self.assertIs(ret_block.jump_edge, None)
        self.assertIs(ret_block.no_jump_edge, None)
        self.assertEqual(len(ret_block.instructions), 1)

        self.assertIs(other_block.jump_edge, None)
        self.assertIs(other_block.no_jump_edge, None)
        self.assertEqual(len(other_block.instructions), 2)

    @unittest.expectedFailure
    def test_jumptables(self):
        lines = '''\
Dump of assembler code for function bar:
   0x0000000000001070 <+0>:	endbr64
   0x0000000000001074 <+4>:	cmp    $0x9,%edi
   0x0000000000001077 <+7>:	ja     0x1120 <bar+176>
   0x000000000000107d <+13>:	lea    0xf80(%rip),%rdx        # 0x2004
   0x0000000000001084 <+20>:	mov    %edi,%edi
   0x0000000000001086 <+22>:	sub    $0x8,%rsp
   0x000000000000108a <+26>:	movslq (%rdx,%rdi,4),%rax
   0x000000000000108e <+30>:	add    %rdx,%rax
   0x0000000000001091 <+33>:	notrack jmpq *%rax
   0x0000000000001094 <+36>:	nopl   0x0(%rax)
   0x0000000000001098 <+40>:	xor    %eax,%eax
   0x000000000000109a <+42>:	callq  0x1040 <foo@plt>
   0x000000000000109f <+47>:	mov    $0xc,%eax
   0x00000000000010a4 <+52>:	add    $0x8,%rsp
   0x00000000000010a8 <+56>:	retq
   0x00000000000010a9 <+57>:	nopl   0x0(%rax)
   0x00000000000010b0 <+64>:	lea    0xf49(%rip),%rdi        # 0x2000
   0x00000000000010b7 <+71>:	xor    %eax,%eax
   0x00000000000010b9 <+73>:	callq  0x1030 <baz@plt>
   0x00000000000010be <+78>:	mov    $0xffffffe7,%eax
   0x00000000000010c3 <+83>:	add    $0x8,%rsp
   0x00000000000010c7 <+87>:	retq
   0x00000000000010c8 <+88>:	nopl   0x0(%rax,%rax,1)
   0x00000000000010d0 <+96>:	mov    $0xffffffe5,%eax
   0x00000000000010d5 <+101>:	add    $0x8,%rsp
   0x00000000000010d9 <+105>:	retq
   0x00000000000010da <+106>:	nopw   0x0(%rax,%rax,1)
   0x00000000000010e0 <+112>:	mov    $0x21,%eax
   0x00000000000010e5 <+117>:	add    $0x8,%rsp
   0x00000000000010e9 <+121>:	retq
   0x00000000000010ea <+122>:	nopw   0x0(%rax,%rax,1)
   0x00000000000010f0 <+128>:	mov    $0x23,%eax
   0x00000000000010f5 <+133>:	add    $0x8,%rsp
   0x00000000000010f9 <+137>:	retq
   0x00000000000010fa <+138>:	nopw   0x0(%rax,%rax,1)
   0x0000000000001100 <+144>:	mov    $0x63,%eax
   0x0000000000001105 <+149>:	add    $0x8,%rsp
   0x0000000000001109 <+153>:	retq
   0x000000000000110a <+154>:	nopw   0x0(%rax,%rax,1)
   0x0000000000001110 <+160>:	mov    $0xa,%eax
   0x0000000000001115 <+165>:	add    $0x8,%rsp
   0x0000000000001119 <+169>:	retq
   0x000000000000111a <+170>:	nopw   0x0(%rax,%rax,1)
   0x0000000000001120 <+176>:	retq
'''.split('\n')
        _, blocks = asm2cfg.parse_lines(lines, False, 'x86')

        # TODO: special block for indirect jumps
        self.assertEqual(len(blocks), 4)

    def test_dummy_block(self):
        lines = '''\
Dump of assembler code for function main:
   0x000055555556fffb <+1709>:	push   %rbx
   0x000055555556fffd <+1707>:	je    0x000055555556fffb <main+0>
'''.split('\n')
        _, blocks = asm2cfg.parse_lines(lines, False, 'x86')

        self.assertEqual(len(blocks), 2)

        source_block = _get_the_source_block(blocks)
        self.assertIsNot(source_block.jump_edge, None)
        self.assertIsNot(source_block.no_jump_edge, None)
        self.assertEqual(len(source_block.instructions), 2)

        fall_block = blocks[source_block.no_jump_edge]
        self.assertIs(fall_block.jump_edge, None)
        self.assertIs(fall_block.no_jump_edge, None)
        self.assertEqual(len(fall_block.instructions), 1)
        self.assertEqual(fall_block.instructions[0].text, 'end of function')

    # TODO:
    # - functions (with and w/o calls)
    # - skip calls



"""
Test cases to test GDB integration.
"""
import subprocess
import os
import re


def test_savecfg_help():
    result = execute_gdb_command('help savecfg')
    assert 'Save an assembly control-flow graph (CFG)' in result.stdout


def test_viewcfg_help():
    result = execute_gdb_command('help viewcfg')
    assert 'Draw an assembly control-flow graph (CFG)' in result.stdout


def test_help_set():
    result = execute_gdb_command('help set')
    assert 'set skipcalls -- Set whether savecfg and viewcfg commands will skip function' in result.stdout


def test_help_set_skipcalls():
    result = execute_gdb_command('help set skipcalls')
    assert 'Set whether savecfg and viewcfg commands will skip function' in result.stdout


def test_help_show_skipcalls():
    result = execute_gdb_command('help show skipcalls')
    assert 'Set whether savecfg and viewcfg commands will skip function' in result.stdout


def test_show_skipcalls():
    result = execute_gdb_command('show skipcalls')
    assert 'Commands savecfg and viewcfg will' in result.stdout


def test_skipcalls_inital_value():
    result = execute_gdb_command('show skipcalls')
    assert parse_option_value(result.stdout) == 'off'


def test_change_skipcalls_value():
    result = execute_gdb_command('set skipcalls')
    assert parse_option_value(result.stdout) == 'on'


def test_savecfg():
    result = execute_gdb_commands(
        ['set confirm off', 'set breakpoint pending on',
         'file test/fixtures/simple_program/hello', 'b main',
         'run', 'savecfg']
    )
    assert os.path.isfile('main.pdf'), result.stdout
    assert 'Saved CFG to a file main.pdf' in result.stdout, result.stdout


def test_viewcfg():
    stdout = ''
    try:
        result = execute_gdb_commands(
            ['set confirm off', 'set breakpoint pending on',
             'file test/fixtures/simple_program/hello', 'b main',
             'run', 'viewcfg']
        )
        stdout = result.stdout
    except subprocess.TimeoutExpired as ex:
        stdout = str(ex.stdout)
    viewcfg_pattern = re.compile(r'Opening a file (.*) with default viewer')
    result = viewcfg_pattern.search(stdout)

    assert result is not None, stdout

    temporary_filename = result.group(1)

    assert os.path.isfile(temporary_filename), temporary_filename


def execute_gdb_command(command):
    return execute_gdb_commands([command])


def execute_gdb_commands(commands):
    project_root_path = os.getcwd()
    os.environ['PYTHONPATH'] = f'{project_root_path}/src/'
    gdb_script_path = f'{project_root_path}/src/gdb_asm2cfg.py'
    gdb_command = ['gdb', '-ex', f'source {gdb_script_path}']
    for command in commands:
        gdb_command.append('-ex')
        gdb_command.append(command)
    gdb_command.append('-ex')
    gdb_command.append('q')
    result = subprocess.run(
        gdb_command, stdout=subprocess.PIPE, stdin=None,
        stderr=None, timeout=2, check=True, universal_newlines=True,
    )
    return result


def parse_option_value(gdb_output):
    output_pattern = re.compile(r'blocks: (on|off)')
    result = output_pattern.search(gdb_output)
    if result:
        return result.group(1)
    return result



"""
Module containing main building blocks to parse assembly and draw CFGs.
"""

import re
import sys
import tempfile

from graphviz import Digraph


# TODO: make this a command-line flag
VERBOSE = 0


def escape(instruction):
    """
    Escape used dot graph characters in given instruction so they will be
    displayed correctly.
    """
    instruction = instruction.replace('<', r'\<')
    instruction = instruction.replace('>', r'\>')
    instruction = instruction.replace('|', r'\|')
    instruction = instruction.replace('{', r'\{')
    instruction = instruction.replace('}', r'\}')
    instruction = instruction.replace('	', ' ')
    return instruction


class BasicBlock:
    """
    Class to represent a node in CFG with straight lines of code without jump
    or calls instructions.
    """

    def __init__(self, key):
        self.key = key
        self.instructions = []
        self.jump_edge = None
        self.no_jump_edge = None

    def add_instruction(self, instruction):
        """
        Add instruction to this block.
        """
        self.instructions.append(instruction)

    def add_jump_edge(self, basic_block_key):
        """
        Add jump target block to this block.
        """
        if isinstance(basic_block_key, BasicBlock):
            self.jump_edge = basic_block_key.key
        else:
            self.jump_edge = basic_block_key

    def add_no_jump_edge(self, basic_block_key):
        """
        Add no jump target block to this block.
        """
        if isinstance(basic_block_key, BasicBlock):
            self.no_jump_edge = basic_block_key.key
        else:
            self.no_jump_edge = basic_block_key

    def get_label(self):
        """
        Return content of the block for dot graph.
        """
        # Left align in dot.
        label = r'\l'.join([escape(i.text) for i in self.instructions])
        # Left justify the last line too.
        label += r'\l'
        if self.jump_edge:
            if self.no_jump_edge:
                label += '|{<s0>No Jump|<s1>Jump}'
            else:
                label += '|{<s1>Jump}'
        return '{' + label + '}'

    def __str__(self):
        return '\n'.join([i.text for i in self.instructions])

    def __repr__(self):
        return '\n'.join([i.text for i in self.instructions])


def print_assembly(basic_blocks):
    """
    Debug function to print the assembly.
    """
    for basic_block in basic_blocks.values():
        print(basic_block)


def read_lines(file_path):
    """ Read lines from the file and return then as a list. """
    lines = []
    with open(file_path, 'r', encoding='utf8') as asm_file:
        lines = asm_file.readlines()
    return lines


# Common regexes
HEX_PATTERN = r'[0-9a-fA-F]+'
HEX_LONG_PATTERN = r'(?:0x0*)?' + HEX_PATTERN


class InputFormat:  # pylint: disable=too-few-public-methods
    """
    An enum which represents various supported input formats
    """
    GDB = 'GDB'
    OBJDUMP = 'OBJDUMP'


def parse_function_header(line):
    """
    Return function name of memory range from the given string line.

    Match lines for non-stripped binaries:
    'Dump of assembler code for function test_function:'
    lines for stripped binaries:
    'Dump of assembler code from 0x555555555faf to 0x555555557008:'
    and lines for obdjdump disassembly:
    '0000000000016bb0 <_obstack_allocated_p@@Base>:'
    """

    objdump_name_pattern = re.compile(fr'{HEX_PATTERN} <([a-zA-Z_0-9@.]+)>:')
    function_name = objdump_name_pattern.search(line)
    if function_name is not None:
        return InputFormat.OBJDUMP, function_name[1]

    function_name_pattern = re.compile(r'function (\w+):$')
    function_name = function_name_pattern.search(line)
    if function_name is not None:
        return InputFormat.GDB, function_name[1]

    memory_range_pattern = re.compile(fr'(?:Address range|from) ({HEX_LONG_PATTERN}) to ({HEX_LONG_PATTERN}):$')
    memory_range = memory_range_pattern.search(line)
    if memory_range is not None:
        return InputFormat.GDB, f'{memory_range[1]}-{memory_range[2]}'

    return None, None


class Address:
    """
    Represents location in program which may be absolute or relative
    """
    def __init__(self, abs_addr, base=None, offset=None):
        self.abs = abs_addr
        self.base = base
        self.offset = offset

    def is_absolute(self):
        return self.base is None

    def is_relative(self):
        return not self.is_absolute()

    def __str__(self):
        if self.offset is not None:
            return f'0x{self.abs:x} ({self.base}+{self.offset})'
        return f'0x{self.abs}'

    def merge(self, other):
        if self.abs is not None:
            assert self.abs is None or self.abs == other.abs
            self.abs = other.abs
        if self.base is not None:
            assert self.base is None or self.base == other.base
            self.base = other.base
        if self.offset is not None:
            assert self.offset is None or self.offset == other.offset
            self.offset = other.offset


class Encoding:
    """
    Represents a sequence of bytes used for instruction encoding
    e.g. the '31 c0' in
    '16bd3:	31 c0                	xor    %eax,%eax'
    """
    def __init__(self, bites):
        self.bites = bites

    def size(self):
        return len(self.bites)

    def __str__(self):
        return ' '.join(map(lambda b: f'{b:#x}', self.bites))


class X86TargetInfo:
    """
    Contains instruction info for X86-compatible targets.
    """

    def __init__(self):
        pass

    def comment(self):
        return '#'

    def is_call(self, instruction):
        # Various flavors of call:
        #   call   *0x26a16(%rip)
        #   call   0x555555555542
        #   addr32 call 0x55555558add0
        return 'call' in instruction.opcode

    def is_jump(self, instruction):
        return instruction.opcode[0] == 'j'

    def is_unconditional_jump(self, instruction):
        return instruction.opcode.startswith('jmp')

    def is_sink(self, instruction):
        """
        Is this an instruction which terminates function execution e.g. return?
        """
        return instruction.opcode.startswith('ret')


class ARMTargetInfo:
    """
    Contains instruction info for ARM-compatible targets.
    """

    def __init__(self):
        pass

    def comment(self):
        return ';'

    def is_call(self, instruction):
        # Various flavors of call:
        #   bl 0x19d90 <_IO_vtable_check>
        # Note that we should be careful to not mix it with conditional
        # branches like 'ble'.
        return instruction.opcode.startswith('bl') \
            and instruction.opcode not in ('blt', 'ble', 'bls')

    def is_jump(self, instruction):
        return instruction.opcode[0] == 'b' and not self.is_call(instruction)

    def is_unconditional_jump(self, instruction):
        return instruction.opcode == 'b'

    def is_sink(self, instruction):
        """
        Is this an instruction which terminates function execution e.g. return?
        Detect various flavors of return like
          bx lr
          pop {r2-r6,pc}
        Note that we do not consider conditional branches (e.g. 'bxle') to sink.
        """
        return re.search(r'\bpop\b.*\bpc\b', instruction.body) \
            or (instruction.opcode == 'bx' and instruction.ops[0] == 'lr') \
            or instruction.opcode == 'udf'


class Instruction:
    """
    Represents a single assembly instruction with it operands, location and
    optional branch target
    """
    def __init__(self, body, text, lineno, address, opcode, ops, target, imm, target_info):  # noqa
        self.body = body
        self.text = text
        self.lineno = lineno
        self.address = address
        self.opcode = opcode
        self.ops = ops
        self.target = target
        self.info = target_info
        if imm is not None and (self.is_jump() or self.is_call()):
            if self.target is None:
                self.target = imm
            else:
                self.target.merge(imm)

    def is_call(self):
        return self.info.is_call(self)

    def is_jump(self):
        return self.info.is_jump(self)

    def is_direct_jump(self):
        return self.is_jump() and re.match(fr'{HEX_LONG_PATTERN}', self.ops[0])

    def is_sink(self):
        return self.info.is_sink(self)

    def is_unconditional_jump(self):
        return self.info.is_unconditional_jump(self)

    def __str__(self):
        result = f'{self.address}: {self.opcode}'
        if self.ops:
            result += f' {self.ops}'
        return result


def parse_address(line):
    """
    Parses leading address of instruction
    """
    address_match = re.match(fr'^\s*(?:0x)?({HEX_PATTERN})\s*(?:<([+-][0-9]+)>)?:(.*)', line)
    if address_match is None:
        return None, line
    address = Address(int(address_match[1], 16), None, int(address_match[2]) if address_match[2] else None)
    return address, address_match[3]


def split_nth(string, count):
    """
    Splits string to equally-sized chunks
    """
    return [string[i:i+count] for i in range(0, len(string), count)]


def parse_encoding(line):
    """
    Parses byte encoding of instruction for objdump disassemblies
    e.g. the '31 c0' in
    '16bd3:	31 c0                	xor    %eax,%eax'
    In addition to X86 supports ARM encoding styles:
    '4:	e1a01000 	mov	r1, r0'
    '50:	f7ff fffe 	bl	0 <__aeabi_dadd>'
    '54:	0002      	movs	r2, r0'
    """
    # Encoding is separated from assembly mnemonic via tab
    # so we allow whitespace separators between bytes
    # to avoid accidentally matching the mnemonic.
    enc_match = re.match(r'^\s*((?:[0-9a-f]{2,8} +)+)(.*)', line)
    if enc_match is None:
        return None, line
    bites = []
    for chunk in enc_match[1].strip().split(' '):
        bites.extend(int(byte, 16) for byte in split_nth(chunk, 2))
    return Encoding(bites), enc_match[2]


def parse_body(line, target_info):
    """
    Parses instruction body (opcode and operands)
    """
    comment_symbol = target_info.comment()
    body_match = re.match(fr'^\s*([^{comment_symbol}<]+)(.*)', line)
    if body_match is None:
        return None, None, None, line
    body = body_match[1].strip()
    line = body_match[2]
    opcode_match = re.match(r'^(\S*)\s*(.*)', body)
    if opcode_match is None:
        return None, None, None, line
    opcode = opcode_match[1]
    ops = opcode_match[2].split(',') if opcode_match[2] else []
    return body, opcode, ops, line


def parse_target(line):
    """
    Parses optional instruction branch target hint
    """
    target_match = re.match(r'\s*<([a-zA-Z_@0-9]+)([+-]0x[0-9a-f]+|[+-][0-9]+)?>(.*)', line)
    if target_match is None:
        return None, line
    offset = target_match[2] or '+0'
    address = Address(None, target_match[1], int(offset, 0))
    return address, target_match[3]


def parse_comment(line, target_info):
    """
    Parses optional instruction comment
    """
    comment_symbol = target_info.comment()
    comment_match = re.match(fr'^\s*{comment_symbol}\s*(.*)', line)
    if comment_match is None:
        return None, line
    comment = comment_match[1]
    imm_match = re.match(fr'^(?:0x)?({HEX_PATTERN})\s*(<.*>)?(.*)', comment)
    if imm_match is None:
        # If no imm was found, ignore the comment.
        # In particular this takes care of useless ARM comments like
        # '82:	46c0      	nop			; (mov r8, r8)'
        return None, ''
    abs_addr = int(imm_match[1], 16)
    if imm_match[2]:
        target, _ = parse_target(imm_match[2])
        target.abs = abs_addr
    else:
        target = Address(abs_addr)
    return target, imm_match[3]


def parse_line(line, lineno, function_name, fmt, target_info):
    """
    Parses a single line of assembly to create Instruction instance
    """

    # Strip GDB prefix and leading whites
    if line.startswith('=> '):
        # Strip GDB marker
        line = line[3:]
    line = line.lstrip()

    address, line = parse_address(line)
    if address is None:
        return None

    if fmt == InputFormat.OBJDUMP:
        encoding, line = parse_encoding(line)
        if not line:
            return encoding

    original_line = line
    body, opcode, ops, line = parse_body(line, target_info)
    if opcode is None:
        return None

    target, line = parse_target(line)

    imm, line = parse_comment(line, target_info)
    if line:
        # Expecting complete parse
        return None

    # Set base symbol for relative addresses
    if address.base is None:
        address.base = function_name
    if target is not None and target.base is None:
        target.base = function_name

    return Instruction(body, original_line, lineno, address, opcode, ops, target, imm, target_info)


class JumpTable:
    """
    Holds info about branch sources and destinations in asm function.
    """

    def __init__(self, instructions):
        # Address where the jump begins and value which address
        # to jump to. This also includes calls.
        self.abs_sources = {}
        self.rel_sources = {}

        # Addresses where jumps end inside the current function.
        self.abs_destinations = set()
        self.rel_destinations = set()

        # Iterate over the lines and collect jump targets and branching points.
        for inst in instructions:
            if inst is None or not inst.is_direct_jump():
                continue

            self.abs_sources[inst.address.abs] = inst.target
            self.abs_destinations.add(inst.target.abs)

            self.rel_sources[inst.address.offset] = inst.target
            self.rel_destinations.add(inst.target.offset)

    def is_destination(self, address):
        if address.abs is not None:
            return address.abs in self.abs_destinations
        if address.offset is not None:
            return address.offset in self.rel_destinations
        return False

    def get_target(self, address):
        if address.abs is not None:
            return self.abs_sources.get(address.abs)
        if address.offset is not None:
            return self.rel_sources.get(address.offset)
        return None


def parse_lines(lines, skip_calls, target_name):  # noqa pylint: disable=unused-argument
    if target_name == 'x86':
        target_info = X86TargetInfo()
    elif target_name == 'arm':
        target_info = ARMTargetInfo()
    else:
        print(f'Unsupported platform {target_name}')
        sys.exit(1)

    instructions = []
    current_function_name = current_format = None
    for num, line in enumerate(lines, 1):
        fmt, function_name = parse_function_header(line)
        if function_name is not None:
            assert current_function_name is None, 'we handle only one function for now'
            if VERBOSE:
                print(f'New function {function_name} (format {fmt})')
            current_function_name = function_name
            current_format = fmt
            continue

        instruction_or_encoding = parse_line(line, num, current_function_name, current_format, target_info)
        if isinstance(instruction_or_encoding, Encoding):
            # Partial encoding for previous instruction, skip it
            continue
        if instruction_or_encoding is not None:
            instructions.append(instruction_or_encoding)
            continue

        if line.startswith('End of assembler dump') or not line:
            continue

        if line.strip() == '':
            continue

        print(f'Unexpected assembly at line {num}:\n  {line}')
        sys.exit(1)

    # Infer target address for jump instructions
    for instruction in instructions:
        if (instruction.target is None or instruction.target.abs is None) \
                and instruction.is_direct_jump():
            if instruction.target is None:
                instruction.target = Address(0)
            instruction.target.abs = int(instruction.ops[0], 16)

    # Infer relative addresses (for objdump or stripped gdb)
    start_address = instructions[0].address.abs
    end_address = instructions[-1].address.abs
    for instruction in instructions:
        for address in (instruction.address, instruction.target):
            if address is not None \
                    and address.offset is None \
                    and start_address <= address.abs <= end_address:
                address.offset = address.abs - start_address

    if VERBOSE:
        print('Instructions:')
        for instruction in instructions:
            if instruction is not None:
                print(f'  {instruction}')

    jump_table = JumpTable(instructions)

    if VERBOSE:
        print('Absolute destinations:')
        for dst in jump_table.abs_destinations:
            print(f'  {dst:#x}')
        print('Relative destinations:')
        for dst in jump_table.rel_destinations:
            print(f'  {dst}')
        print('Absolute branches:')
        for src, dst in jump_table.abs_sources.items():
            print(f'  {src:#x} -> {dst}')
        print('Relative branches:')
        for src, dst in jump_table.rel_sources.items():
            print(f'  {src} -> {dst}')

    # Now iterate over the assembly again and split it to basic blocks using
    # the branching information from earlier.
    basic_blocks = {}
    current_basic_block = None
    previous_jump_block = None
    for line, instruction in zip(lines, instructions):
        if instruction is None:
            continue

        # Current offset/address inside the function.
        program_point = instruction.address
        jump_point = jump_table.get_target(program_point)
        is_unconditional = instruction.is_unconditional_jump()

        if current_basic_block is None:
            current_basic_block = BasicBlock(program_point.abs)
            basic_blocks[current_basic_block.key] = current_basic_block
            # Previous basic block ended in jump instruction. Add the basic
            # block what follows if the jump was not taken.
            if previous_jump_block is not None:
                previous_jump_block.add_no_jump_edge(current_basic_block)
                previous_jump_block = None
        elif jump_table.is_destination(program_point):
            temp_block = current_basic_block
            current_basic_block = BasicBlock(program_point.abs)
            basic_blocks[current_basic_block.key] = current_basic_block
            temp_block.add_no_jump_edge(current_basic_block)

        current_basic_block.add_instruction(instruction)

        if jump_point is not None:
            current_basic_block.add_jump_edge(jump_point.abs)
            previous_jump_block = None if is_unconditional else current_basic_block
            current_basic_block = None
        elif instruction.is_sink():
            previous_jump_block = current_basic_block = None

    if previous_jump_block is not None:
        # If last instruction of the function is jump/call, then add dummy
        # block to designate end of the function.
        end_block = BasicBlock('end_of_function')
        dummy_instruction = Instruction('', 'end of function', 0, None, None, [], None, None, target_info)
        end_block.add_instruction(dummy_instruction)
        previous_jump_block.add_no_jump_edge(end_block.key)
        basic_blocks[end_block.key] = end_block

    return current_function_name, basic_blocks


def draw_cfg(function_name, basic_blocks, view):
    dot = Digraph(name=function_name, comment=function_name, engine='dot')
    dot.attr('graph', label=function_name)
    for address, basic_block in basic_blocks.items():
        key = str(address)
        dot.node(key, shape='record', label=basic_block.get_label())
    for basic_block in basic_blocks.values():
        if basic_block.jump_edge:
            if basic_block.no_jump_edge is not None:
                dot.edge(f'{basic_block.key}:s0', str(basic_block.no_jump_edge))
            dot.edge(f'{basic_block.key}:s1', str(basic_block.jump_edge))
        elif basic_block.no_jump_edge:
            dot.edge(str(basic_block.key), str(basic_block.no_jump_edge))
    if view:
        dot.format = 'gv'
        with tempfile.NamedTemporaryFile(mode='w+b', prefix=function_name) as filename:
            dot.view(filename.name)
            print(f'Opening a file {filename.name}.{dot.format} with default viewer. Don\'t forget to delete it later.')
    else:
        dot.format = 'pdf'
        dot.render(filename=function_name, cleanup=True)
        print(f'Saved CFG to a file {function_name}.{dot.format}')



"""
Command-line usage support.
"""

import argparse
from . import asm2cfg


def main():
    """ Command-line entry point to the program. """
    parser = argparse.ArgumentParser(
        description='Program to draw dot control-flow graph from GDB disassembly for a function.',
        epilog='If function CFG rendering takes too long, try to skip function calls with -c flag.'
    )
    parser.add_argument('assembly_file',
                        help='File to contain one function assembly dump')
    parser.add_argument('-c', '--skip-calls', action='store_true',
                        help='Skip function calls from dividing code to blocks')
    parser.add_argument('--target', choices=['x86', 'arm'], default='x86',
                        help='Specify target platform for assembly')
    parser.add_argument('-v', '--view', action='store_true',
                        help='View as a dot graph instead of saving to a file')
    args = parser.parse_args()
    print('If function CFG rendering takes too long, try to skip function calls with -c flag')
    lines = asm2cfg.read_lines(args.assembly_file)
    function_name, basic_blocks = asm2cfg.parse_lines(lines, args.skip_calls, args.target)
    asm2cfg.draw_cfg(function_name, basic_blocks, args.view)
